@@ -1,10 +1,11 @@
 #!/usr/bin/env perl
 
 # Flex(1) XML processor scanner generator.
-# Copyright  1999 Kristoffer Rose.  All rights reserved.
+# Copyright (C) 1999 Kristoffer Rose.  All rights reserved.
 #
 # This file is part of the FleXML XML processor generator system.
-# Copyright  1999 Kristoffer Rose.  All rights reserved.
+# Copyright (C) 1999 Kristoffer Rose.  All rights reserved.
+# Copyright (C) 2003-2011 Martin Quinson.  All rights reserved.
 #
 # This program is free software; you can redistribute it and/or modify it
 # under the terms of the GNU General Public License as published by the Free
@@ -22,12 +23,12 @@
 
 use warnings;
 
-my $Id = '$Id: flexml.pl,v 1.48 2006/03/21 13:09:12 mquinson Exp $ ';
+my $Id = '$Id: flexml.pl,v 1.67 2011/11/01 20:49:21 mquinson Exp $ ';
 $Id =~ s/\s*\$\s*//go;
 
 # IMPORTS.
 
-use Getopt::Std;
+use Getopt::Long;
 use LWP::UserAgent;
 use Carp qw/cluck confess/;
 
@@ -41,10 +42,11 @@ use vars qw/ $HEADER /;		# generated XML processor API header
 use vars qw/ $ACTIONS /;	# the actions file
 use vars qw/ $APPLICATION /;	# generated XML application
 
+$SKELETON = "./skel"; # default value
+
 # OPTIONS (and other globals).
 
 my $Use;			# usage string
-my %opt = ();			# options
 
 my $debug;			# -d option flag
 my $verbose;			# -v option flag
@@ -53,9 +55,19 @@ my $nofail;			# -X option flag
 my $quiet_parser;		# -q option flag
 my $uri;			# -u option uri
 my $pubid;			# -p option string
-my $stacksize;			# -b option flag
-my $tagprefix;			# -P option flag
-my $actbin;                     # -T option content
+my $sysid;			# --sysid option string
+my $stacksize=100000;		# -b option flag
+my $tagprefix="";      		# -P option flag
+my $actbin="./flexml-act";      # -T option content
+my $init_header="";             # -i option string
+
+my $header;			# -H option flag/content
+my $dummy;			# -D option flag/content
+my $standalone;			# -A option flag/content
+my $scanner;			# -S option flag/content
+my $actions;			# -a option content
+
+my $dryrun;                     # -n option flag
 
 my $dtd;			# DTD file name (or URI)
 my $dtdrevision;		# DTD version pruned from file
@@ -83,6 +95,7 @@ my @tags = ();		# Tags of the DTD.
 my %roottags = ();	# Tags that may be the root tag.
 my %ctag = ();		# C variable name of each tag.
 
+my %allstates = ();     # all existing states are key of this hash
 my %states = ();	# $states{tag} is list of states used by tag element.
 my %emptytrans = ();	# $emptytrans{state} contains empty transitions in automaton.
 
@@ -166,7 +179,8 @@ sub redistribute { # Print C comment with generated file "license".
   my ($pre) = @_;
   print <<EOT;
 $pre This program was generated with the FleXML XML processor generator.
-$pre FleXML is Copyright  1999-2005 Kristoffer Rose.  All rights reserved.
+$pre FleXML is Copyright (C) 1999-2005 Kristoffer Rose.  All rights reserved.
+$pre FleXML is Copyright (C) 2003-2011 Martin Quinson.  All rights reserved.
 $pre ($Id).
 $pre 
 $pre There are two, intertwined parts to this program, part A and part B.
@@ -175,8 +189,8 @@ $pre Part A
 $pre ------
 $pre 
 $pre Some parts, here collectively called "Part A", are found in the 
-$pre FleXML package.  They are Copyright  1999-2005 Kristoffer Rose. 
-$pre All rights reserved.
+$pre FleXML package.  They are Copyright (C) 1999-2005 Kristoffer Rose
+$pre and Copyright (C) 2003-2011 Martin Quinson. All rights reserved.
 $pre
 $pre You can redistribute, use, perform, display and/or modify "Part A"
 $pre provided the following two conditions hold:
@@ -234,12 +248,23 @@ sub api_types { # Print XML application interface types.
 
 
 sub api_data { # Print XML application interface parameters.
-  my ($pre) = @_;
+  my ($pre) = @_; # pre should be a storage class spec like 'static' or 'extern'
   print "/* FleXML-provided data. */\n";
-  print $pre . "const char* ${tagprefix}pcdata;\n";
+  print $pre . "int ${tagprefix}pcdata_ix;\n";
+  print "extern char *${tagprefix}bufferstack;\n";
+  print "#define ${tagprefix}pcdata (${tagprefix}bufferstack + ${tagprefix}pcdata_ix)\n";
+
   for (keys %atttype) {
     if (m.($Nmtoken)[/]($Nmtoken).xo) {
-      print $pre . "AT_${tagprefix}$ctag{$1}_$catt{$2} A_${tagprefix}$ctag{$1}_$catt{$2};\n";
+      print $pre . "AT_${tagprefix}$ctag{$1}_$catt{$2} AX_${tagprefix}$ctag{$1}_$catt{$2};\n";
+      print "#define A_${tagprefix}$ctag{$1}_$catt{$2} ";
+      if ($enumtype{$_}) {
+	  print "AX_${tagprefix}$ctag{$1}_$catt{$2}\n";
+      }
+      else {
+	  print "(${tagprefix}bufferstack + AX_${tagprefix}$ctag{$1}_$catt{$2})\n";
+      }
+      print $pre . "short int ${tagprefix}$ctag{$1}_$catt{$2}_isset;\n";
     }
   }
 }
@@ -248,7 +273,7 @@ sub api_data { # Print XML application interface parameters.
 sub expandparametersat { # Expand parameter entities in $_ at $place.
   my ($place) = @_;
   while ( m/$place\s*%($Name);/ ) {
-    if ($parameter{$1}) {
+    if (defined $parameter{$1}) {
       s/$place(\s*)%($Name);/ ' '. $1 . $parameter{$2} . ' ' /xe;
     }
   }
@@ -300,6 +325,16 @@ sub geturl { # Insert contents of URL into input stream at current point
 #  }
 }
 
+# add attribute to list of all default attributes
+my @default_attributes;
+my $next_att_loc = 1;
+sub add_def_att {
+    my ($att) = @_;
+    my $retval = $next_att_loc;
+    push @default_attributes, ($att);
+    $next_att_loc += length($att) + 1;
+    return $retval;
+}
 
 sub nextline {			# return one input line
 
@@ -392,8 +427,10 @@ sub analysechildren { # Analyse DTD children specification; return
                       # true if it may be empty.  Uses global $statecounter.
 
   my ($tag,$re,$in,$out) = @_;
+  $allstates{$in} = 1;
+  $allstates{$out} = 1;
 
-  #print "analysechildren [ $tag, $re, $in, $out ] \n";
+  print "analysechildren [ $tag, $re, $in, $out ] \n" if $debug;
 
   local $_ = $re;
 
@@ -487,49 +524,71 @@ sub analysechildren { # Analyse DTD children specification; return
 
 
 # Parse options.
-$Use = "Usage: flexml [-ASHDvdqnLXV] [-s skel] [-T actbin] [-p pubid] [-u uri]\n"
-     . "       [-b stack_size] [-r roottags] [-a actions] [-P prefix] name[.dtd]";
+$Use = "Usage: flexml [-ASHDvdqnLXV] [-s skel] [-T actbin] [--sysid sysid] "
+  . "[-p pubid] [-u uri] [-i init_header]\n"
+  . "       [-b stack_increment] [-r roottags] [-a actions] [-P prefix] name[.dtd]";
 
-getopts('ASHDvdnLXVqp:b:P:s:T:u:r:a:', \%opt);
+sub show_version {
 
-# Version!
-print "FleXML version $Id.\n" if $opt{V} or $opt{v};
-exit 0 if $opt{V};
-
-# Debugging?
-$debug = $opt{d};
-$verbose = $opt{v};
-
-# Line numbers?
-$lineno = $opt{L};
-
-# Quiet parser?    
-$quiet_parser = $opt{q};
-    
-# Exit without fail message?
-$nofail = $opt{X};
-
-# Specific root tags?
-if ($opt{r}) {
-  for (split ',',$opt{r}) { $roottags{$_} = 'true'; }
+    exit 0;
 }
-
-# Specific stack size?
-$stacksize = ($opt{'b'} ? $opt{'b'} : 100000);
-
-# Specific tagprefix?
-$tagprefix = ($opt{'P'} ? $opt{'P'}."_" : "");
-
-# Specific actbin?
-$actbin = ($opt{'T'} ? $opt{'T'} : "./flexml-act");
     
-# Set skeleton scanner file name and check it is readable (if needed).
-$SKELETON = ($opt{'s'} ? $opt{'s'} : './skel');
-die "$0: No skeleton file $SKELETON.\n" if not -r $SKELETON and $opt{S};
+Getopt::Long::Configure ("bundling");
+GetOptions(
+    # Debugging and verbosity
+    "debug|d" => \$debug,
+    "verbose|v" => \$verbose,
+    "quiet|q" => \$quiet_parser,
+    
+    # Version!
+    "version|V" => sub { print "FleXML version $Id.\n"; exit 0; },
 
-# Set document type URI and PUBID.
-$uri = $opt{u} if $opt{u};
-$pubid = $opt{p} if $opt{p};
+    # dry-run ?
+    "dry-run|n" => \$dryrun,
+
+    # Line numbers?
+    "lineno|L" => \$lineno,
+    
+    # Exit without fail message?
+    "nofail|X" => \$nofail,
+
+    # Specific root tags?
+    "root-tags|r=s" => sub {
+	for (split ',',$_[1]) { $roottags{$_} = 'true'; }
+    },
+
+    # Specific stack size?
+    "stack-increment|b=s" => \$stacksize,
+    
+    # Specific tagprefix?
+    "tag-prefix|P=s" => sub { $tagprefix = $_[1]."_" },
+    
+    # Specific actbin? (internal use)
+    "act-bin|T=s" => \$actbin,
+    
+    # Set skeleton scanner file name and check it is readable (if needed).
+    "skel|s=s" => sub {
+	$SKELETON = $_[1];
+	die "$0: No skeleton file $SKELETON.\n" if not -r $SKELETON and $_[1];
+    },
+    
+    # Set document type URI and PUBID.
+    "uri|u=s" => \$uri,
+    "pubid|p=s" => \$pubid,
+    "sysid=s" => \$sysid,
+
+    # name of header file to be included in initial section of generated .l file
+    "init_header|i=s" => \$init_header,
+
+    # What to generate
+    "header|H:s" => sub { $header = $_[1] || 'true' },
+    "dummy|D:s" => sub { $dummy = $_[1] || 'true' },
+    "stand-alone|A" => \$standalone,
+    "scanner|S:s" => sub { $scanner = $_[1] || 'true' },
+    "actions|a=s" => \$actions
+    );
+
+print "FleXML version $Id.\n" if $verbose;
 
 # Set DTD file name...and extract prefix for later
 my $prefix = $ARGV[0];
@@ -546,30 +605,30 @@ else {
 
 # Selection options: If none of -SHDA specified then default to -SH.
 # Furthermore -a implies -D.
-$opt{S} = $opt{H} = 'true' unless ($opt{S} or $opt{H} or $opt{D} or $opt{A});
-$opt{D} ||= $opt{a} unless $opt{A};
+$scanner = $header = 'true' unless ($scanner or $header or $dummy or $standalone);
+$dummy ||= $actions unless $standalone;
 
 # Set default (DTD-based) output file names.
-$SCANNER = "$prefix.l";
-$HEADER =  "$prefix.h";
-$APPLICATION = "$prefix-dummy.c";
+$SCANNER =     (!defined($scanner)) || $scanner eq 'true' ? "$prefix.l"       : $scanner;
+$HEADER =      (!defined($header))  || $header eq 'true'  ? "$prefix.h"       : $header;
+$APPLICATION = (!defined($dummy))   || $dummy eq 'true'   ? "$prefix-dummy.c" : $dummy;
 
 # Set actions=based output file names, if any.
-if ($ACTIONS = $opt{a}) {
-  $opt{a} =~ s/\.[a-z]+$//;
-  $APPLICATION = "$opt{a}.c";
+if ($ACTIONS = $actions) {
+  $actions =~ s/\.[a-z]+$//;
+  $APPLICATION = "$actions.c";
 }
 
 # Stand-alone applications...
-if ($opt{A}) {
-  die "$0: -A conflicts with -SHD.\n" if ($opt{S} or $opt{H} or $opt{D});
+if ($standalone) {
+  die "$0: -A conflicts with -SHD.\n" if ($scanner or $header or $dummy);
   $SCANNER = $APPLICATION;
   $SCANNER =~ s/\.c$/.l/;
 }
 
 # Dry-run?
-if ($opt{n}) {
-  $opt{A} = $opt{S} = $opt{H} = $opt{D} = undef;
+if ($dryrun) {
+  $standalone = $scanner = $header = $dummy = undef;
 }
 
 
@@ -672,11 +731,12 @@ while (@inputs) {		# While there are lines...
     # strip off the matched code from the beginning of $_
     $_ = $'; #'
 
+    # place the newly found tag into the list of tags @tags
+    die place()."Repeated element $tag.\n" if exists $source{$tag};
+
     # ????? - most of the time $orig is undef
     $source{$tag} = "$orig";
 
-    # place the newly found tag into the list of tags @tags
-    die place()."Repeated element $tag.\n" if $ctag{$tag};
     push @tags, $tag;
 
     # Create C-friendly tag names.
@@ -815,22 +875,22 @@ while (@inputs) {		# While there are lines...
 	  }
 	  # ...and string/token types.
 	  else {
-	    $typeof{"$tag/$attribute"} = 'const char*';
+	    $typeof{"$tag/$attribute"} = 'int';
 	    if ($default) {
-	      $initof{"$tag/$attribute"} = "\"$default\"";
+	      $initof{"$tag/$attribute"} = add_def_att($default);
 	    }
 	    else {
-	      $initof{"$tag/$attribute"} = 'NULL';
+	      $initof{"$tag/$attribute"} = '0';
 	    }
 
 	    # Special treatment of token types.
 	    if ( $type eq 'ID' or $type eq 'IDREF' ) {
 	      $literaltype{"$tag/$attribute"} = '{Name}';
-	      print STDERR place()."Warning: attribute type `$type' not validated.\n";
+	      print STDERR place()."Warning: cannot validate attribute type `$type' (yet).\n";
 	    }
 	    elsif ( $type eq 'IDREFS' ) {
 	      $literaltype{"$tag/$attribute"} = '{Names}';
-	      print STDERR place()."Warning: attribute type `$type' not validated.\n";
+	      print STDERR place()."Warning: cannot validate attribute type `$type' (yet).\n";
 	    }
 	    elsif ( $type eq 'NMTOKEN' ) {
 	      $literaltype{"$tag/$attribute"} = '{Nmtoken}';
@@ -839,10 +899,15 @@ while (@inputs) {		# While there are lines...
 	      $literaltype{"$tag/$attribute"} = '{Nmtokens}';
 	    }
 	    elsif ( $type eq 'ENTITY' ) {
-	      die place()."ENTITY attribute type unimplemented.\n";
+	      #die place()."ENTITY attribute type unimplemented.\n";
+	      $literaltype{"$tag/$attribute"} = '{Name}';
+	      print STDERR place()."Warning: cannote validate attribute type `$type' (yet).\n";
 	    }
 	    elsif ( $type eq 'ENTITIES' ) {
-	      die place()."ENTITIES attribute type unimplemented.\n";
+	      #die place()."ENTITIES attribute type unimplemented.\n";
+	      $literaltype{"$tag/$attribute"} = '{Names}';
+	      print STDERR place()."Warning: cannot validate attribute type `$type' (yet).\n";
+
 	    }
 	    elsif ( $type ne 'CDATA' ) {
 	      die place()."Unknown AttType `$type'.\n";
@@ -858,7 +923,7 @@ while (@inputs) {		# While there are lines...
     }
 
     $orig = '';			# in case there were no attributes...
-  }
+  } # end of a <!ATTLIST> block
 
   # Parse internal parameter entity declaration.
   elsif ( m/^\s*<!ENTITY\s+%\s+($Name)\s+'([^'']*)'\s*>\s*/xo
@@ -980,7 +1045,7 @@ if (not %roottags) {
 }
 
 # Handling prefix
-if($opt{P}) {
+if(length($tagprefix)) {
     my($h,$k);
     my(@hashlist) = (\%states,\%emptytrans,\%instates,\%startstate,
 		     \%endstates,\%exittrans);
@@ -995,6 +1060,13 @@ if($opt{P}) {
 	    $$h{$k} =~ s/>S_/>S_${tagprefix}/g;
 	}
     }
+    my %tmp = ();
+    foreach $k (keys %allstates) {
+	$k =~ s/^E_/E_${tagprefix}/;
+	$k =~ s/^S_/S_${tagprefix}/;
+	$tmp{$k} = 1;
+    }
+    %allstates = %tmp;
 }
 
 # Debugging: show DTD representation.
@@ -1016,7 +1088,32 @@ if ($debug) {
     return $out;
   }
 
-  print '%opt = (' . printhash(\%opt) . ")\n";
+  # display the options we got
+  print "debug=".($debug||'undef')."\n";
+  print "verbose=".($verbose||'undef')."\n";
+  print "quiet=".($quiet_parser||'undef')."\n";
+
+  print "dry-run=".($dryrun||'undef')."\n";
+  print "lineno=".($lineno||'undef')."\n";
+
+  print "nofail=".($nofail||'undef')."\n";
+
+  print "stack-increment=".($stacksize||'undef')."\n";
+  print "tag-prefix=".($tagprefix||'undef')."\n";
+
+  print "act-bin=".($actbin||'undef')."\n";
+
+  print "skel=".($SKELETON||'undef')."\n";
+
+  print "uri=".($uri||'undef')."\n";
+  print "pubid=".($pubid||'undef')."\n\n";
+  print "sysid=".($sysid||'undef')."\n\n";
+
+  print "header=".($header||'undef')."\n";
+  print "dummy=".($dummy||'undef')."\n";
+  print "stand-alone=".($standalone||'undef')."\n";
+  print "scanner=".($scanner||'undef')."\n";
+  print "actions=".($actions||'undef')."\n\n";
 
   print '%source = (' . printhash(\%source) . ")\n";
   print "\n";
@@ -1065,7 +1162,7 @@ if ($debug) {
 
 # WRITE API HEADER (if requested).
 
-if ($opt{H}) {
+if ($header) {
 
   print STDOUT "Generating XML processor header in `$HEADER'.\n" if $verbose;
 
@@ -1112,10 +1209,10 @@ if ($opt{H}) {
 # WRITE XML PROCESSOR (if requested).
 
 
-if ($opt{S} or $opt{A}) {
+if ($scanner or $standalone) {
 
   print STDOUT "Writing XML processor"
-             . ($opt{a} || $opt{A} ? " and application" : "")
+             . ($actions || $standalone ? " and application" : "")
              . " onto `$SCANNER'.\n" if $verbose;
 
   open SCANNER, "+>$SCANNER"|| die "$0: cannot write $SCANNER: $!\n";
@@ -1154,11 +1251,14 @@ if ($opt{S} or $opt{A}) {
       print "#define FLEXML_yylineno\n" if $lineno;
       print "#define FLEXML_NOFAIL\n" if $nofail;
       print "#define FLEXML_quiet_parser\n" if $quiet_parser;
-      print "#define FLEXML_HasMixed\n" if %inmixed;
+      print "#ifndef FLEXML_BUFFERSTACKSIZE\n";
       print "#define FLEXML_BUFFERSTACKSIZE $stacksize\n";
+      print "#endif\n";
+      print "#define FLEXML_NEED_BUFFERLIT\n"
+	if (scalar(%literaltype) or ($#default_attributes >= 0));
       print "\n";
 
-      if ($opt{A}) {
+      if ($standalone) {
 	api_functions('static ',';');
 	print "\n";
 	api_types();
@@ -1168,11 +1268,18 @@ if ($opt{S} or $opt{A}) {
       else {
 	print "/* XML processor api. */\n";
 	print "#include \"$HEADER\"\n\n"
-	    if ($opt{H});
+	    if ($header);
 	api_data('');
       }
 
     }
+    elsif ( /^FLEXML_INCLUDE_INIT_HEADER$/ ) {
+	if ($init_header) {
+	    print "/* User-supplied header */\n";
+	    print "#include \"$init_header\"\n";
+	}
+    }
+
     elsif ( /^FLEXML_FLEX_OPTIONS$/ ) {
 
       print "%option yylineno\n" if $lineno;
@@ -1203,7 +1310,10 @@ if ($opt{S} or $opt{A}) {
       my ($state, $tag);
 
       print "  /* FleXML_init */\n";
-      print "  next = bufferstack;\n";
+      print "  bnext = inext = 1;\n";
+      for my $att (@default_attributes) {
+	  print "  ${tagprefix}bufferliteral('\\0', &bnext, \"$att\");\n";
+      }
       print "  if(!${tagprefix}statenames) {${tagprefix}statenames= (const char **)calloc(IMPOSSIBLE,sizeof(char*));\n";
       for ('PROLOG','DOCTYPE','EPILOG','INCOMMENT','INPI','VALUE1','VALUE2','CDATA') {
 	print "  ${tagprefix}statenames[$_] = NULL;\n";
@@ -1220,8 +1330,9 @@ if ($opt{S} or $opt{A}) {
     }
     elsif ( /^FLEXML_DOCTYPES$/ ) {
 
-      my $sysid = ( $uri ? "(\"'$uri'\"|\"\\\"$uri\\\"\")"
-			    : "(\"'$dtd'\"|\"\\\"$dtd\\\"\")" );
+      $sysid = $sysid ?          "(\"'$sysid'\"|\"\\\"$sysid\\\"\")"
+	              : ( $uri ? "(\"'$uri'\"|\"\\\"$uri\\\"\")"
+			       : "(\"'$dtd'\"|\"\\\"$dtd\\\"\")" );
 
       for (keys %roottags) {
 	my $c = $ctag{$_};
@@ -1268,12 +1379,24 @@ if ($opt{S} or $opt{A}) {
 	  }
 	  $exitswitch .= "  }\n";
 	}
+
+	# Misplaced start or empty tag
+	unless ($nofail) {
+	    print "\n";
+	    my %ins = ();
+	    foreach (keys %allstates) {
+		$ins{$_} = 'false';
+	    }
+	    if (exists $instates{$tag}) {
+		for (split /,/,$instates{$tag}) { $ins{$_} = 'true'; }
+	    }
+	    my $str = '<' . join(',', grep {$ins{$_} eq 'false'} keys %ins);	    
+	    if ($str ne '<') {
+		print "$str>\"<$tag\"{s} FAIL(\"Starting tag <$tag> is not allowed here.\");\n";
+	    }
+	}
 	
-
-
-
-
-
+	  
 	# Start or empty tag: initialise attribute list.
 	print "\n";
 	if ($roottags{$tag}) {
@@ -1288,9 +1411,10 @@ if ($opt{S} or $opt{A}) {
 	print ">\"<$tag\"{s} {\n";
 
 	for my $attribute (@myattributes) {
-	  print "  A_${tagprefix}${myctag}_$catt{$attribute} = " . $initof{"$tag/$attribute"} . ";\n";
+	  print "  AX_${tagprefix}${myctag}_$catt{$attribute} = " . $initof{"$tag/$attribute"} . ";\n";
+	  print "  ${tagprefix}${myctag}_$catt{$attribute}_isset = 0;\n";
 	}
-	print "  ENTER(AL_${tagprefix}$myctag);\n";
+	print "  ENTER(AL_${tagprefix}$myctag); pushbuffer(0);\n";
 	print "  }\n";
 
 #	print " .       FAIL(\"Unexpected character `%c': `<$tag' expected.\",yytext[0]);\n";
@@ -1333,23 +1457,23 @@ if ($opt{S} or $opt{A}) {
 	    print " \"$attribute\"{Eq}\"'" . $attdef{"$tag/$attribute"} . "'\""
 	        . " |\n"
 		. " \"$attribute\"{Eq}\"\\\"" . $attdef{"$tag/$attribute"} . "\\\"\""
-		. " A_${tagprefix}${myctag}_$catt{$attribute}"
+		. " AX_${tagprefix}${myctag}_$catt{$attribute}"
 		. " = " . $initof{"$tag/$attribute"} . ";\n";
 
 	  }
 	  elsif ($type = $literaltype{"$tag/$attribute"}) {
 
 	    # - (non-fixed) literal-type attribute: scan literal string directly, or
-	    print " \"$attribute\"{Eq}\'$type\' BUFFERLITERAL('\\\'',A_${tagprefix}${myctag}_$catt{$attribute});\n";
+	    print " \"$attribute\"{Eq}\'$type\' BUFFERLITERAL('\\\'',AX_${tagprefix}${myctag}_$catt{$attribute});\n";
 
-	    print " \"$attribute\"{Eq}\\\"$type\\\" BUFFERLITERAL('\"',A_${tagprefix}${myctag}_$catt{$attribute});\n";
+	    print " \"$attribute\"{Eq}\\\"$type\\\" BUFFERLITERAL('\"',AX_${tagprefix}${myctag}_$catt{$attribute});\n";
 
 	  }
 	  else {
 
 	    # - (non-fixed non-literal) attribute: scan string with entity expansion.
-	    print " \"$attribute\"{Eq}\\' ENTER(VALUE1); BUFFERSET(A_${tagprefix}${myctag}_$catt{$attribute});\n";
-	    print " \"$attribute\"{Eq}\\\" ENTER(VALUE2); BUFFERSET(A_${tagprefix}${myctag}_$catt{$attribute});\n";
+	    print " \"$attribute\"{Eq}\\' if (${tagprefix}${myctag}_$catt{$attribute}_isset != 0) {FAIL(\"Multiple definition of attribute $attribute in <${tagprefix}${myctag}>\");} ${tagprefix}${myctag}_$catt{$attribute}_isset = 1; ENTER(VALUE1); BUFFERSET(AX_${tagprefix}${myctag}_$catt{$attribute});\n";
+	    print " \"$attribute\"{Eq}\\\" if (${tagprefix}${myctag}_$catt{$attribute}_isset != 0) {FAIL(\"Multiple definition of attribute $attribute in <${tagprefix}${myctag}>\");}  ${tagprefix}${myctag}_$catt{$attribute}_isset = 1; ENTER(VALUE2); BUFFERSET(AX_${tagprefix}${myctag}_$catt{$attribute});\n";
 
 	  }
 	  print "\n";
@@ -1360,13 +1484,13 @@ if ($opt{S} or $opt{A}) {
 	print " \">\" {\n";
 	for my $attribute (@myattributes) {
 	  if ($required{"$tag/$attribute"}) {
-	    print "  if (!A_${tagprefix}$ctag{$tag}_" . variablify($attribute) . ")"
+	    print "  if (!AX_${tagprefix}$ctag{$tag}_" . variablify($attribute) . ")"
 	        . " FAIL(\"Required attribute `$attribute' not set for `$tag' element.\");\n";
 	  }
 	}
 	print "  LEAVE; STag_${tagprefix}$myctag();"
-	    . (%inmixed ? ' pushbuffer('."${tagprefix}".'pcdata);' : '')
-	    . ($mixed{$tag} ? 'BUFFERSET('."${tagprefix}".'pcdata)' : "${tagprefix}".'pcdata = NULL'). ";"
+	    . (%inmixed ? ' pushbuffer('."${tagprefix}".'pcdata_ix);' : '')
+	    . ($mixed{$tag} ? 'pushbuffer('."${tagprefix}".'pcdata_ix); BUFFERSET('."${tagprefix}".'pcdata_ix);' : "${tagprefix}".'pcdata_ix = 0'). ";"
 	    . " ENTER($startstate{$tag});\n";
 	print " }\n";
 	#
@@ -1375,16 +1499,16 @@ if ($opt{S} or $opt{A}) {
 	  print " \"/>\" {\n";
 	  for my $attribute (@myattributes) {
 	    if ($required{"$tag/$attribute"}) {
-	      print "  if (!A_${tagprefix}$ctag{$tag}_" . variablify($attribute) . ")"
+	      print "  if (!AX_${tagprefix}$ctag{$tag}_" . variablify($attribute) . ")"
 		  . " FAIL(\"Required attribute `$attribute' not set for `$tag' element.\");\n";
 	    }
 	  }
 	  print "  LEAVE; STag_${tagprefix}$myctag();"
-	      . (%inmixed ? ' pushbuffer('."${tagprefix}".'pcdata);' : '')
-	      . " ${tagprefix}".'pcdata = ' . ($mixed{$tag} ? '""' : 'NULL') . ';'
+	      . (%inmixed ? ' pushbuffer('."${tagprefix}".'pcdata_ix);' : '')
+	      . " ${tagprefix}".'pcdata_ix = 0;'
 	      . " ETag_${tagprefix}$myctag();"
-	      . (%inmixed ? " ${tagprefix}".'pcdata = popbuffer();' : '')
-	      . "\n";
+	      . (%inmixed ? " ${tagprefix}".'pcdata_ix = popbuffer();' : '')
+	      . " popbuffer(); /* attribute */\n";
 	  #
 	  print $exitswitch;
 	  print " }\n";
@@ -1408,7 +1532,9 @@ if ($opt{S} or $opt{A}) {
 	print "  LEAVE;\n";
 	print "  BUFFERDONE;\n" if $mixed{$tag};
 	print "  ETag_${tagprefix}$myctag();\n";
-	print "  ${tagprefix}pcdata = popbuffer();\n" if %inmixed;
+	print "  ${tagprefix}pcdata_ix = popbuffer();\n" if $mixed{$tag};
+	print "  ${tagprefix}pcdata_ix = popbuffer();\n" if %inmixed;
+	print "  popbuffer(); /* attribute */\n";
 	print $exitswitch;
 	print " }\n";
 
@@ -1505,13 +1631,15 @@ if ($opt{S} or $opt{A}) {
       s/element_context/${tagprefix}element_context/g;
       s/parse_err_msg/${tagprefix}parse_err_msg/g;
       s/rcs_flexml_skeleton/rcs_${tagprefix}flexml_skeleton/g;
+      s/bufferliteral/${tagprefix}bufferliteral/g;
+      s/bufferstack/${tagprefix}bufferstack/g;
       print;
     }
 
   }
 
   close SKELETON || die "$0: Cannot close $SKELETON: $!\n";
-  unless ($opt{A}) {
+  unless ($standalone) {
     close SCANNER || die "$0: Cannot close $SCANNER: $!\n";
   }
 }
@@ -1521,10 +1649,10 @@ if ($opt{S} or $opt{A}) {
 # WRITE APPLICATION.
 
 
-if ($opt{D}) {
+if ($dummy) {
 
   print STDOUT "Writing XML"
-             . ($opt{a} ? "" : " dummy")
+             . ($actions ? "" : " dummy")
              . " application onto `$APPLICATION'.\n" if $verbose;
 
   open APPLICATION, "+>$APPLICATION" || die "$0: Cannot write $APPLICATION: $!\n";
@@ -1546,7 +1674,7 @@ if ($opt{D}) {
 
 }
 
-if ($opt{D} or $opt{A}) {
+if ($dummy or $standalone) {
 
   # Get requested actions.
   if ($ACTIONS) {
@@ -1589,23 +1717,17 @@ if ($opt{D} or $opt{A}) {
       if ($tag) {
 
 	while ( s/\{($Name)\}/A_${tagprefix}$ctag{$tag}_$catt{$1}/x ) {
-	  die "\"$ACTIONS\", line $lineno: Attributes only allowed in start tag.\n"
-	    if not $isstart;
 	  die "\"$ACTIONS\", line $lineno: Unknown attribute `$1' for <$tag>.\n"
 	    if not $atttype{"$tag/$1"};
 	}
 
 	while ( s/\{[!]($Name)\}/AU_${tagprefix}$ctag{$tag}_$catt{$1}/x ) {
-	  die "\"$ACTIONS\", line $lineno: Attributes only allowed in start tag.\n"
-	    if not $isstart;
 	  die "\"$ACTIONS\", line $lineno: Unknown attribute `$1' for <$tag>.\n"
 	    if not $atttype{"$tag/$1"};
 	}
 
 	while ( s|\{($Name)=($Name)\}|
 		  "A_${tagprefix}$ctag{$tag}_$catt{$1}_" . variablify($2); |xe ) {
-	  die "\"$ACTIONS\", line $lineno: Attributes only allowed in start tag.\n"
-	    if not $isstart;
 	  my ($att,$elt) = ($1,$2);
 	  die "\"$ACTIONS\", line $lineno: Unknown attribute $1 for <$tag>.\n"
 	    if not $atttype{"$tag/$1"};
@@ -1623,6 +1745,7 @@ if ($opt{D} or $opt{A}) {
 	die "\"$ACTIONS\", line $lineno: Malformed annotation `$&' in <$tag> action.\n"
 	  if m|\{[^;\s]+\}|o;
       }
+      print STDERR "Action: $_" if $debug;
       print $_;
 
     }
@@ -1635,10 +1758,10 @@ if ($opt{D} or $opt{A}) {
 
 }
 
-if ($opt{D}) {
+if ($dummy) {
   close APPLICATION || die "$0: Cannot close $APPLICATION: $!\n";
 }
-elsif ($opt{A}) {
+elsif ($standalone) {
   close SCANNER || die "$0: Cannot close $SCANNER: $!\n";
 }
 
@@ -1656,6 +1779,7 @@ B<flexml>
 [B<-ASHDvdnLXV>]
 [B<-s>I<skel>]
 [B<-p>I<pubid>]
+[B<-i>I<init_header>]
 [B<-u>I<uri>]
 [B<-r>I<rootags>]
 [B<-a>I<actions>]
@@ -1700,102 +1824,114 @@ I<Flexml> takes the following options.
 
 =over 4
 
-=item B<-A>
+=item B<--stand-alone>, B<-A> 
 
 Generate a I<stand-alone> scanner application.  If combined with
 B<-a>I<actions> then the application will be named as I<actions> with
 the extension replaced by F<.l>, otherwise it will be in I<name>F<.l>.
 Conflicts with B<-S>, B<-H>, and B<-D>.
 
-=item B<-a> I<actions>
+=item B<--actions> I<actions>, B<-a> I<actions>
 
 Uses the I<actions> file to produce an XML application in the file
 with the same name as I<actions> after replacing the extension with
 F<.c>.  If combined with B<-A> then instead the stand-alone
 application will include the action functions.
 
-=item B<-D>
+=item B<--dummy> B<[>I<app_name>B<]>, B<-D> B<[>I<app_name>B<]>
 
-Generate a dummy application I<name>F<-dummy.c> with just empty
-functions to be called by the XML processor.  If combined with
-B<-a>I<actions> then the application will insert the specified
-actions and be named as I<actions> with the extension replaced by
-F<.c>.  Conflicts with B<-A>; implied by B<-a> unless either of
-B<-SHD> is specified.
+Generate a dummy application with just empty functions to be called by the
+XML processor. If I<app_name> is not specified on the command line, it
+defaults to I<name>F<-dummy.c>.  If combined with B<-a> I<actions> then the
+application will insert the specified actions and be named as I<actions>
+with the extension replaced by F<.c>.  Conflicts with B<-A>; implied by
+B<-a> unless either of B<-SHD> is specified.
 
-=item B<-d>
+=item B<--debug>, B<-d>
 
 Turns on debug mode in the flex scanner and also prints out the
 details of the DTD analysis performed by I<flexml>.
 
-=item B<-H>
+=item B<--header> B<[>I<header_name>B<]>, B<-H> B<[>I<header_name>B<]>
 
-Generate the header file I<name>F<.h>.  Conflicts with B<-A>; on by
+Generate the header file. If the I<header_name> is not specified on the
+command line, defaults to I<name>F<.h>.  Conflicts with B<-A>; on by
 default if none of B<-SHD> specified.
 
-=item B<-L>
+=item B<--lineno>, B<-L>
 
 Makes the XML processor (as produced by I<flex>(1)) count the lines in
 the input and keep it available to XML application actions in the
 integer C<yylineno>.  (This is off by default as the performance
 overhead is significant.)
 
-=item B<-q>
+=item B<--quiet>, B<-q>
 
 Prevents the XML processor (as produced by I<flex>(1)) from reporting 
 the error it runs into on stderr. Instead, users will have to pool for
 error messages with the parse_err_msg() function.
 By default, error messages are written on stderr.
 
-=item B<-n>
+=item B<--dry-run>, B<-n>
 
 "Dry-run": do not produce any of the output files.
 
-=item B<-p> I<pubid>
+=item B<--pubid> I<pubid>, B<-p> I<pubid>
 
 Sets the document type to be C<PUBLIC> with the identifier I<pubid>
 instead of C<SYSTEM>, the default.
 
-=item B<-r> I<roottags>
+=item B<--init_header> I<init_header>, B<-i> I<init_header>
+
+Puts a line containing C<#include "init_header"> in the C<%{...%}> section
+at the top of the generated .l file.  This may be useful for making various
+flex C<#define>s, for example C<YY_INPUT> or C<YY_DECL>.
+
+=item B<--sysid>=I<sysid>
+
+Overrides the C<SYSTEM> id of the accepted DTD. Sometimes useful when your
+dtd is placed in a subdirectory.
+
+=item B<--root-tags> I<roottags>, B<-r> I<roottags>
 
 Restricts the XML processor to validate only documents with one of the
 root elements listed in the comma-separated I<roottags>.
 
-=item B<-S>
+=item B<--scanner> B<[>I<scanner_name>B<]>, B<-S> B<[>I<scanner_name>B<]>
 
-Generate the scanner I<name>F<.l>.  Conflicts with B<-A>; on by
-default if none of B<-SHD> specified.
+Generate the scanner. If I<scanner_name> is not given on command line, it
+defaults to I<name>F<.l>.  Conflicts with B<-A>; on by default if none of
+B<-SHD> specified.
 
-=item B<-s> I<skel>
+=item B<--skel> I<skel>, B<-s> I<skel>
 
 Use the skeleton scanner I<skel> instead of the default.
 
-=item B<-T> I<flexml-act>
+=item B<--act-bin> I<flexml-act>, B<-T> I<flexml-act>
 
 This is an internal option mainly used to test versions of flexml not
 installed yet.
 
-=item B<-b> I<stack_size>
+=item B<--stack-increment> I<stack_increment>, B<-b> I<stack_increment>
 
-Sets the FLEXML_BUFFERSTACKSIZE to stack_size (100000 by default). Use
-this option when you get an error like "Assertion `next<limit'
-failed".
+Sets the FLEXML_BUFFERSTACKSIZE to stack_increment (100000 by default). This
+controls how much the data stack grows in each realloc().
 
-=item B<-p> I<STRING>
+=item B<--tag-prefix> I<STRING>, B<-O> I<STRING>
 
 Use STRING to differentiate multiple versions of flexml in the same C
-code.
+code, just like the -P flex argument.
 
-=item B<-u> I<uri>
+=item B<--uri> I<uri>, B<-u> I<uri>
 
 Sets the URI of the DTD, used in the C<DOCTYPE> header, to the
 specified I<uri> (the default is the DTD name).
 
-=item B<-v>
+=item B<--verbose>, B<-v>
 
 Be verbose: echo each DTD declaration (after parameter expansion).
 
-=item B<-V>
+=item B<--version>, B<-V>
 
 Print the version of I<flexml> and exit.
 
@@ -1946,11 +2082,6 @@ in particular C<xml:spaces> should be supported.
 
 =item *
 
-The XML processor currently uses a fixed-size buffer to read
-C<pcdata>.  It should not.
-
-=item *
-
 The DTD parser is presently a perl hack so it may parse some DTDs
 badly; in particular the expansion of parameter entities may not
 conform fully to the XML specification.
@@ -1984,7 +2115,7 @@ REC-xml-1998-0210).
 
 =head1 AUTHOR
 
-I<Flexml> was written by Kristoffer Hgsbro Rose,
+I<Flexml> was written by Kristoffer Rose,
 E<lt>C<krisrose@debian.org>E<gt>.
 
 =head1 COPYRIGHT
